@@ -69,7 +69,7 @@ fn main() -> openpgp::Result<()> {
         Default::default();
     let mut p_comp: HashMap<Vec<CompressionAlgorithm>, usize> =
         Default::default();
-    let mut p_aead: HashMap<Vec<AEADAlgorithm>, usize> =
+    let mut p_aead_ciphersuites: HashMap<Vec<(SymmetricAlgorithm, AEADAlgorithm)>, usize> =
         Default::default();
 
     // Per-Cert statistics.
@@ -80,8 +80,10 @@ fn main() -> openpgp::Result<()> {
 
     // UserAttribute statistics.
     let mut ua_image_count = vec![0; 256];
+    let mut ua_image_unknown_variant_count = 0;
     let mut ua_unknown_count = vec![0; 256];
     let mut ua_invalid_count = 0;
+    let mut ua_unknown_variant_count = 0;
 
     // Key statistics.
     let mut pk_algo_size: HashMap<PublicKeyAlgorithm, HashMap<usize, usize>> =
@@ -95,7 +97,7 @@ fn main() -> openpgp::Result<()> {
 
     // Current certificate.
     let mut current_fingerprint =
-        KeyHandle::Fingerprint(Fingerprint::from_bytes(&vec![0; 20]));
+        KeyHandle::Fingerprint(Fingerprint::from_bytes(4, &vec![0; 20])?);
     let mut current_keyid = KeyHandle::KeyID(KeyID::wildcard());
 
     // For each input file, create a parser.
@@ -210,12 +212,12 @@ fn main() -> openpgp::Result<()> {
                                     } else {
                                         p_comp.insert(a.clone(), 1);
                                     },
-                                SubpacketValue::PreferredAEADAlgorithms(a)
+                                SubpacketValue::PreferredAEADCiphersuites(a)
                                     =>
-                                    if let Some(count) = p_aead.get_mut(a) {
+                                    if let Some(count) = p_aead_ciphersuites.get_mut(a) {
                                         *count += 1;
                                     } else {
-                                        p_aead.insert(a.clone(), 1);
+                                        p_aead_ciphersuites.insert(a.clone(), 1);
                                     },
                                 SubpacketValue::ExportableCertification(v) =>
                                     if *v {
@@ -250,9 +252,13 @@ fn main() -> openpgp::Result<()> {
                                     ua_image_count[n as usize] += 1,
                                 Image::Unknown(n, _) =>
                                     ua_image_count[n as usize] += 1,
+                                _ =>
+                                    ua_image_unknown_variant_count += 1,
                             },
                             Ok(Subpacket::Unknown(n, _)) =>
                                 ua_unknown_count[n as usize] += 1,
+                            Ok(_) =>
+                                ua_unknown_variant_count += 1,
                             Err(_) => ua_invalid_count += 1,
                         }
                     }
@@ -528,15 +534,15 @@ fn main() -> openpgp::Result<()> {
         }
     }
 
-    if !p_aead.is_empty() {
+    if !p_aead_ciphersuites.is_empty() {
         println!();
-        println!("# PreferredAEADAlgorithms statistics");
+        println!("# PreferredAEADCiphersuites statistics");
         println!();
         println!("{:>70} {:>9}", "", "count",);
         println!("----------------------------------------\
                   ----------------------------------------");
 
-        for (a, n) in p_aead.iter() {
+        for (a, n) in p_aead_ciphersuites.iter() {
             let a = format!("{:?}", a);
             println!("{:>70} {:>9}", &a[1..a.len()-1], n);
         }
@@ -549,13 +555,13 @@ fn main() -> openpgp::Result<()> {
         println!();
         println!("# User Attribute Subpacket statistics");
         println!();
-        println!("{:>18} {:>9}",
+        println!("{:>21} {:>9}",
                  "", "count",);
         println!("----------------------------");
         for t in 0..256 {
             let n = ua_image_count[t];
             if n > 0 {
-                println!("{:>18} {:>9}",
+                println!("{:>21} {:>9}",
                          match t {
                              1 =>         "Image::JPEG".into(),
                              100..=110 => format!("Image::Private({})", t),
@@ -563,14 +569,22 @@ fn main() -> openpgp::Result<()> {
                          }, n);
             }
         }
+        if ua_image_unknown_variant_count > 0 {
+            println!("{:>21} {:>9}", "Unknown image variant",
+                     ua_image_unknown_variant_count);
+        }
         for t in 0..256 {
             let n = ua_unknown_count[t];
             if n > 0 {
-                println!("{:>18} {:>9}", format!("Unknown({})", t), n);
+                println!("{:>21} {:>9}", format!("Unknown({})", t), n);
             }
         }
+        if ua_unknown_variant_count > 0 {
+            println!("{:>21} {:>9}", "Unknown variant",
+                     ua_unknown_variant_count);
+        }
         if ua_invalid_count > 0 {
-            println!("{:>18} {:>9}", "Invalid", ua_invalid_count);
+            println!("{:>21} {:>9}", "Invalid", ua_invalid_count);
         }
     }
 
