@@ -1,7 +1,6 @@
 use std::io;
 use std::mem;
 use std::vec;
-use std::path::Path;
 
 use buffered_reader::BufferedReader;
 use lalrpop_util::ParseError;
@@ -344,7 +343,7 @@ impl CertValidator {
 /// this way, it is possible to propagate parse errors.
 ///
 /// A `CertParser` returns each [`TPK`] or [`TSK`] that it encounters.
-/// Note: if you don't actually need all of the certificates, it is
+/// Note: if you don't actually need all the certificates, it is
 /// usually faster to use a [`RawCertParser`] and only fully parse and
 /// canonicalize those certificates that are relevant.
 ///
@@ -404,8 +403,8 @@ impl CertValidator {
 /// goal is to provide some future compatibility.
 ///
 /// [`Packet`]: crate::packet::Packet
-/// [`TPK`]: https://tools.ietf.org/html/rfc4880#section-11.1
-/// [`TSK`]: https://tools.ietf.org/html/rfc4880#section-11.2
+/// [`TPK`]: https://www.rfc-editor.org/rfc/rfc9580.html#section-10.1
+/// [`TSK`]: https://www.rfc-editor.org/rfc/rfc9580.html#section-10.2
 /// [`Public Key`]: super::Packet::PublicKey
 /// [`Secret Key`]: super::Packet::SecretKey
 /// [`Literal Data`]: super::Packet::Literal
@@ -427,10 +426,10 @@ impl CertValidator {
 ///
 /// # fn main() -> Result<()> {
 /// # let (alice, _) =
-/// #       CertBuilder::general_purpose(None, Some("alice@example.org"))
+/// #       CertBuilder::general_purpose(Some("alice@example.org"))
 /// #       .generate()?;
 /// # let (bob, _) =
-/// #       CertBuilder::general_purpose(None, Some("bob@example.org"))
+/// #       CertBuilder::general_purpose(Some("bob@example.org"))
 /// #       .generate()?;
 /// #
 /// # let mut keyring = Vec::new();
@@ -471,21 +470,21 @@ impl CertValidator {
 /// use openpgp::types::DataFormat;
 ///
 /// # fn main() -> Result<()> {
-/// let mut lit = Literal::new(DataFormat::Text);
+/// let mut lit = Literal::new(DataFormat::Unicode);
 /// lit.set_body(b"test".to_vec());
 ///
 /// let (alice, _) =
-///       CertBuilder::general_purpose(None, Some("alice@example.org"))
+///       CertBuilder::general_purpose(Some("alice@example.org"))
 ///       .generate()?;
 /// let (bob, _) =
-///       CertBuilder::general_purpose(None, Some("bob@example.org"))
+///       CertBuilder::general_purpose(Some("bob@example.org"))
 ///       .generate()?;
 ///
 /// let mut packets : Vec<Packet> = Vec::new();
-/// packets.extend(alice.clone());
+/// packets.extend(alice.clone().into_packets());
 /// packets.push(lit.clone().into());
 /// packets.push(lit.clone().into());
-/// packets.extend(bob.clone());
+/// packets.extend(bob.clone().into_packets());
 ///
 /// let r : Vec<Result<Cert>> = CertParser::from(packets).collect();
 /// assert_eq!(r.len(), 4);
@@ -601,25 +600,12 @@ impl<'a> Parse<'a, CertParser<'a>> for CertParser<'a>
     where
         R: BufferedReader<Cookie> + 'a,
     {
-        Ok(Self::from(PacketParser::from_buffered_reader(reader)?))
-    }
-
-    /// Initializes a `CertParser` from a `Read`er.
-    fn from_reader<R: 'a + io::Read + Send + Sync>(reader: R) -> Result<Self> {
-        Ok(Self::from(PacketParser::from_reader(reader)?))
-    }
-
-    /// Initializes a `CertParser` from a `File`.
-    fn from_file<P: AsRef<Path>>(path: P) -> Result<Self> {
-        Ok(Self::from(PacketParser::from_file(path)?))
-    }
-
-    /// Initializes a `CertParser` from a byte string.
-    fn from_bytes<D: AsRef<[u8]> + ?Sized + Send + Sync>(data: &'a D) -> Result<Self> {
-        Ok(Self::from(PacketParser::from_bytes(data)?))
+        Ok(Self::from(PacketParser::from_buffered_reader(reader.into_boxed())?))
     }
 }
 
+impl<'a> crate::seal::Sealed for CertParser<'a> {}
+
 impl<'a> CertParser<'a> {
     /// Creates a `CertParser` from a `Result<Packet>` iterator.
     ///
@@ -642,10 +628,10 @@ impl<'a> CertParser<'a> {
     ///
     /// # fn main() -> Result<()> {
     /// # let (alice, _) =
-    /// #       CertBuilder::general_purpose(None, Some("alice@example.org"))
+    /// #       CertBuilder::general_purpose(Some("alice@example.org"))
     /// #       .generate()?;
     /// # let (bob, _) =
-    /// #       CertBuilder::general_purpose(None, Some("bob@example.org"))
+    /// #       CertBuilder::general_purpose(Some("bob@example.org"))
     /// #       .generate()?;
     /// #
     /// # let mut keyring = Vec::new();
@@ -694,7 +680,7 @@ impl<'a> CertParser<'a> {
     /// certificates in a large keyring, most certificates can be
     /// immediately discarded.  That is, it is more efficient to
     /// filter, validate, and double check, than to validate and
-    /// filter.  (It is necessary to double check, because the check
+    /// filter.  (It is necessary to double-check, because the check
     /// might have been on an invalid part.  For example, if searching
     /// for a key with a particular Key ID, a matching key might not
     /// have any self signatures.)
@@ -766,11 +752,26 @@ impl<'a> CertParser<'a> {
     // returns None.
     fn parse(&mut self, p: Packet) -> Result<Option<Cert>> {
         tracer!(TRACE, "CertParser::parse", 0);
-        if let Packet::Marker(_) = p {
-            // Ignore Marker Packet.  RFC4880, section 5.8:
-            //
-            //   Such a packet MUST be ignored when received.
-            return Ok(None);
+        match p {
+            Packet::Marker(_) => {
+                // Ignore Marker Packet.  RFC4880, section 5.8:
+                //
+                //   Such a packet MUST be ignored when received.
+                return Ok(None);
+            },
+            Packet::Padding(_) => {
+                // "[Padding packets] MUST be ignored when received.",
+                // section 5.14 of RFC 9580.
+                return Ok(None);
+            },
+            p if Cert::valid_packet(&p).is_err()
+                && ! p.tag().is_critical() =>
+            {
+                // "Unknown, non-critical packets MUST be ignored when
+                // received.", section 4.3 of RFC 9580.
+                return Ok(None);
+            },
+            _ => {},
         }
 
         if !self.packets.is_empty() {
@@ -853,7 +854,7 @@ impl<'a> CertParser<'a> {
         if failed {
             // There was at least one packet that doesn't belong in a
             // Cert.  Fail now.
-            let err = Error::UnsupportedCert2(
+            let err = Error::UnsupportedCert(
                 "Packet sequence includes non-Cert packets.".into(),
                 packets);
             t!("Invalid certificate: {}", err);
@@ -883,7 +884,7 @@ impl<'a> CertParser<'a> {
         }).and_then(|mut cert| {
             let primary_fp: KeyHandle = cert.key_handle();
 
-            // The parser puts all of the signatures on the
+            // The parser puts all the signatures on the
             // certifications field.  Split them now.
 
             split_sigs(&primary_fp, &mut cert.primary);
@@ -936,7 +937,7 @@ fn split_sigs<C>(primary: &KeyHandle, b: &mut ComponentBundle<C>)
             || issuers.iter().any(|kh| kh.aliases(primary));
 
         use crate::SignatureType::*;
-        if typ == AttestationKey {
+        if typ == CertificationApproval {
             b.attestations.push(sig);
         } else if typ == KeyRevocation
             || typ == SubkeyRevocation
@@ -1311,7 +1312,8 @@ mod test {
 
         fn cert_cmp(a: &Result<Cert>, b: &Vec<Packet>)
         {
-            let a : Vec<Packet> = a.as_ref().unwrap().clone().into();
+            let a =
+                a.as_ref().unwrap().clone().into_packets().collect::<Vec<_>>();
 
             for (i, (a, b)) in a.iter().zip(b).enumerate() {
                 if a != b {
@@ -1326,9 +1328,9 @@ mod test {
         }
 
         let (cert, _) =
-            CertBuilder::general_purpose(None, Some("alice@example.org"))
+            CertBuilder::general_purpose(Some("alice@example.org"))
             .generate()?;
-        let cert : Vec<Packet> = cert.into();
+        let cert = cert.into_packets().collect::<Vec<_>>();
 
         // A userid packet.
         let userid : Packet = cert.clone()
@@ -1344,7 +1346,7 @@ mod test {
             .into();
 
         // A literal packet.  (This is a valid OpenPGP Message.)
-        let mut lit = Literal::new(DataFormat::Text);
+        let mut lit = Literal::new(DataFormat::Unicode);
         lit.set_body(b"test".to_vec());
         let lit = Packet::from(lit);
 
@@ -1586,14 +1588,14 @@ mod test {
         for i in 0..n {
             let (cert, _) =
                 CertBuilder::general_purpose(
-                    None, Some(format!("{}@example.org", i)))
+                    Some(format!("{}@example.org", i)))
                 .generate()?;
 
             cert.as_tsk().serialize(&mut data)?;
             certs_orig.push(cert);
 
             if literal {
-                let mut lit = Literal::new(DataFormat::Text);
+                let mut lit = Literal::new(DataFormat::Unicode);
                 lit.set_body(b"data".to_vec());
 
                 Packet::from(lit).serialize(&mut data)?;
@@ -1810,14 +1812,14 @@ mod test {
 
         let (cert_1, _) =
             CertBuilder::general_purpose(
-                None, Some("a@example.org"))
+                Some("ea@example.org"))
             .generate()?;
         let cert_1_packets: Vec<Packet>
-            = cert_1.into_packets2().collect();
+            = cert_1.into_packets().collect();
 
         let (cert_2, _) =
             CertBuilder::general_purpose(
-                None, Some("b@example.org"))
+                Some("eb@example.org"))
             .generate()?;
 
         for n in 1..cert_1_packets.len() {
