@@ -68,12 +68,9 @@
 use std::borrow::Cow;
 use std::convert::TryFrom;
 use std::fmt;
-use std::io::Read;
-use std::path::Path;
 
-use buffered_reader::{BufferedReader, Dup, EOF, File, Generic, Memory};
+use buffered_reader::{BufferedReader, Dup, EOF, Memory};
 
-use crate::Error;
 use crate::Fingerprint;
 use crate::KeyID;
 use crate::Result;
@@ -147,7 +144,7 @@ pub use iter::KeyIter;
 /// #     Ok(())
 /// # }
 /// ```
-#[derive(Clone, PartialEq, Eq)]
+#[derive(Clone, PartialEq, Eq, Hash, PartialOrd, Ord)]
 pub struct RawPacket<'a> {
     tag: Tag,
     header_len: usize,
@@ -208,14 +205,14 @@ impl<'a> crate::serialize::Marshal for RawPacket<'a> {
 /// A mostly unparsed `Cert`.
 ///
 /// This data structure contains the unparsed packets for a
-/// certificate or key.  The packet sequence is well formed in the
+/// certificate or key.  The packet sequence is well-formed in the
 /// sense that the sequence of tags conforms to the [Transferable
 /// Public Key grammar] or [Transferable Secret Key grammar], and that
 /// it can extract the primary key's fingerprint.  Beyond that, the
 /// packets are not guaranteed to be valid.
 ///
-/// [Transferable Public Key grammar]: https://www.rfc-editor.org/rfc/rfc4880#section-11.1
-/// [Transferable Secret Key grammar]: https://www.rfc-editor.org/rfc/rfc4880#section-11.2
+/// [Transferable Public Key grammar]: https://www.rfc-editor.org/rfc/rfc9580.html#section-10.1
+/// [Transferable Secret Key grammar]: https://www.rfc-editor.org/rfc/rfc9580.html#section-10.2
 ///
 /// This data structure exists to quickly split a large keyring, and
 /// only parse those certificates that appear to be relevant.
@@ -223,7 +220,7 @@ impl<'a> crate::serialize::Marshal for RawPacket<'a> {
 pub struct RawCert<'a> {
     data: Cow<'a, [u8]>,
 
-    fingerprint: Fingerprint,
+    primary_key: Key<key::PublicParts, key::PrimaryRole>,
 
     // The packet's tag, the length of the header, and the offset of
     // the start of the packet (including the header) into data.
@@ -234,7 +231,7 @@ assert_send_and_sync!(RawCert<'_>);
 impl<'a> fmt::Debug for RawCert<'a> {
     fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
         f.debug_struct("RawCert")
-            .field("fingerprint", &self.fingerprint)
+            .field("fingerprint", &self.fingerprint())
             .field("packets",
                    &self.packets
                    .iter()
@@ -267,12 +264,12 @@ impl<'a> RawCert<'a> {
 
     /// Returns the certificate's fingerprint.
     pub fn fingerprint(&self) -> Fingerprint {
-        self.fingerprint.clone()
+        self.primary_key.fingerprint()
     }
 
     /// Returns the certificate's Key ID.
     pub fn keyid(&self) -> KeyID {
-        KeyID::from(&self.fingerprint)
+        KeyID::from(self.fingerprint())
     }
 
     /// Returns the ith packet.
@@ -384,19 +381,15 @@ impl<'a> RawCert<'a> {
     fn keys_internal(&self)
         -> impl Iterator<Item=Key<key::PublicParts, key::UnspecifiedRole>> + '_
     {
-        self.packets()
-            .filter_map(|p| {
-                if matches!(p.tag(),
-                            Tag::PublicKey | Tag::PublicSubkey
-                            | Tag::SecretKey | Tag::SecretSubkey)
-                {
-                    Key::from_bytes(p.body())
-                        .ok()
-                        .map(|k| k.parts_into_public())
-                } else {
-                    None
-                }
-            })
+        std::iter::once(self.primary_key().clone().role_into_unspecified())
+            .chain(self.packets()
+                   .filter(|p| matches!(p.tag(),
+                                        Tag::PublicKey | Tag::PublicSubkey
+                                        | Tag::SecretKey | Tag::SecretSubkey))
+                   .skip(1) // The primary key.
+                   .filter_map(|p| Key::from_bytes(p.body())
+                               .ok()
+                               .map(|k| k.parts_into_public())))
     }
 
     /// Returns the certificate's primary key.
@@ -410,7 +403,7 @@ impl<'a> RawCert<'a> {
     ///
     /// [`ValidCert`]: crate::cert::ValidCert
     pub fn primary_key(&self) -> Key<key::PublicParts, key::PrimaryRole> {
-        self.keys().next().expect("have a primary key").role_into_primary()
+        self.primary_key.clone()
     }
 
     /// Returns the certificate's User IDs.
@@ -435,6 +428,62 @@ impl<'a> RawCert<'a> {
                 }
             })
     }
+
+    /// Changes the `RawCert`'s lifetime to the static lifetime.
+    ///
+    /// Returns a `RawCert` with a static lifetime by copying any
+    /// referenced data.
+    ///
+    /// [`RawCertParser::next`] returns a `RawCert` with the same
+    /// lifetime as its reader.  In certain situations,
+    /// `RawCertParser::next` can take advantage of this to avoid
+    /// copying data.  Tying the `RawCert`'s lifetime to the reader is
+    /// inconvenient when the `RawCert` needs to outlive the reader,
+    /// however.  This function copies any referenced data thereby
+    /// breaking the dependency.
+    ///
+    /// ```
+    /// # use sequoia_openpgp::Result;
+    /// # use sequoia_openpgp::cert::raw::RawCert;
+    /// # use sequoia_openpgp::cert::raw::RawCertParser;
+    /// # use sequoia_openpgp::parse::Parse;
+    ///
+    /// # fn main() -> Result<()> {
+    /// fn read_certs<'a>() -> Result<Vec<RawCert<'static>>> {
+    ///     let input = // ...
+    ///     # Vec::new();
+    ///
+    ///     // The lifetime of the returned certs is tied to input.
+    ///     // We use into_owned to break the dependency.
+    ///     let parser = RawCertParser::from_bytes(&input)?;
+    ///     let certs = parser
+    ///         .map(|r| r.map(|c| c.into_owned()))
+    ///         .collect::<Result<Vec<_>>>()?;
+    ///     Ok(certs)
+    /// }
+    ///
+    /// let cert = read_certs()?;
+    /// # assert_eq!(cert.len(), 0);
+    /// # Ok(()) }
+    /// ```
+    pub fn into_owned(self) -> RawCert<'static> {
+        match self.data {
+            Cow::Owned(data) => {
+                RawCert {
+                    data: Cow::Owned(data),
+                    primary_key: self.primary_key,
+                    packets: self.packets,
+                }
+            }
+            Cow::Borrowed(data) => {
+                RawCert {
+                    data: Cow::Owned(data.to_vec()),
+                    primary_key: self.primary_key,
+                    packets: self.packets,
+                }
+            }
+        }
+    }
 }
 
 impl<'a> TryFrom<&RawCert<'a>> for Cert {
@@ -461,26 +510,22 @@ impl<'a> Parse<'a, RawCert<'a>> for RawCert<'a> {
     where
         R: BufferedReader<Cookie> + 'a
     {
-        let mut parser = RawCertParser::from_buffered_reader(reader)?;
-        if let Some(cert_result) = parser.next() {
-            if parser.next().is_some() {
-                Err(Error::MalformedCert(
-                    "Additional packets found, is this a keyring?".into()
-                ).into())
+        fn parse<'a>(reader: Box<dyn BufferedReader<Cookie> + 'a>) -> Result<RawCert<'a>> {
+            let mut parser = RawCertParser::from_buffered_reader(reader)?;
+            if let Some(cert_result) = parser.next() {
+                if parser.next().is_some() {
+                    Err(crate::Error::MalformedCert(
+                        "Additional packets found, is this a keyring?".into()
+                    ).into())
+                } else {
+                    cert_result
+                }
             } else {
-                cert_result
+                Err(crate::Error::MalformedCert("No data".into()).into())
             }
-        } else {
-            Err(Error::MalformedCert("No data".into()).into())
         }
-    }
 
-    /// Returns the first RawCert encountered in the reader.
-    ///
-    /// Returns an error if there are multiple certificates.
-    fn from_reader<R: 'a + Read + Send + Sync>(reader: R) -> Result<Self> {
-        let br = Generic::with_cookie(reader, None, Cookie::default());
-        Self::from_buffered_reader(br)
+        parse(reader.into_boxed())
     }
 }
 
@@ -506,14 +551,14 @@ impl<'a> crate::serialize::Marshal for RawCert<'a> {
 ///
 /// [`CertParser`]: crate::cert::CertParser
 ///
-/// `RawCertParser` checks that the packet sequence is well formed in
+/// `RawCertParser` checks that the packet sequence is well-formed in
 /// the sense that the sequence of tags conforms to the [Transferable
 /// Public Key grammar] or [Transferable Secret Key grammar], and it
 /// performs a few basic checks.  See the documentation for
 /// [`RawCert`] for details.
 ///
-/// [Transferable Public Key grammar]: https://www.rfc-editor.org/rfc/rfc4880#section-11.1
-/// [Transferable Secret Key grammar]: https://www.rfc-editor.org/rfc/rfc4880#section-11.2
+/// [Transferable Public Key grammar]: https://www.rfc-editor.org/rfc/rfc9580.html#section-10.1
+/// [Transferable Secret Key grammar]: https://www.rfc-editor.org/rfc/rfc9580.html#section-10.2
 ///
 /// Because a `RawCertParser` doesn't parse the contents of the
 /// packets, it is significantly faster than a [`CertParser`] when
@@ -592,8 +637,7 @@ pub struct RawCertParser<'a>
 assert_send_and_sync!(RawCertParser<'_>);
 
 impl<'a> RawCertParser<'a> {
-    fn new<R>(reader: R) -> Result<Self>
-    where R: 'a + BufferedReader<Cookie>
+    fn new(reader: Box<dyn BufferedReader<Cookie> + 'a>) -> Result<Self>
     {
         // Check that we can read the first header and that it is
         // reasonable.  Note: an empty keyring is not an error; we're
@@ -606,7 +650,7 @@ impl<'a> RawCertParser<'a> {
                 Ok(header) => {
                     let tag = header.ctb().tag();
                     if matches!(tag, Tag::Unknown(_) | Tag::Private(_)) {
-                        return Err(Error::MalformedCert(
+                        return Err(crate::Error::MalformedCert(
                             format!("A certificate must start with a \
                                      public key or a secret key packet, \
                                      got a {}",
@@ -635,7 +679,7 @@ impl<'a> RawCertParser<'a> {
                 Ok(header) => {
                     let tag = header.ctb().tag();
                     if matches!(tag, Tag::Unknown(_) | Tag::Private(_)) {
-                        return Err(Error::MalformedCert(
+                        return Err(crate::Error::MalformedCert(
                             format!("A certificate must start with a \
                                      public key or a secret key packet, \
                                      got a {}",
@@ -669,23 +713,14 @@ impl<'a> Parse<'a, RawCertParser<'a>> for RawCertParser<'a>
     where
         R: BufferedReader<Cookie> + 'a
     {
-        RawCertParser::new(reader)
-    }
-
-    /// Initializes a `RawCertParser` from a `Read`er.
-    fn from_reader<R: 'a + Read + Send + Sync>(reader: R) -> Result<Self> {
-        RawCertParser::new(Generic::with_cookie(reader, None, Default::default()))
-    }
-
-    /// Initializes a `RawCertParser` from a `File`.
-    fn from_file<P: AsRef<Path>>(path: P) -> Result<Self> {
-        RawCertParser::new(File::with_cookie(path, Default::default())?)
+        RawCertParser::new(reader.into_boxed())
     }
 
     /// Initializes a `RawCertParser` from a byte string.
     fn from_bytes<D: AsRef<[u8]> + ?Sized + Send + Sync>(data: &'a D) -> Result<Self> {
         let data = data.as_ref();
-        let mut p = RawCertParser::new(Memory::with_cookie(data, Default::default()))?;
+        let mut p = RawCertParser::new(
+            Memory::with_cookie(data, Default::default()).into_boxed())?;
 
         // If we are dearmoring the input, then the slice doesn't
         // reflect the raw packets.
@@ -696,6 +731,8 @@ impl<'a> Parse<'a, RawCertParser<'a>> for RawCertParser<'a>
     }
 }
 
+impl<'a> crate::seal::Sealed for RawCertParser<'a> {}
+
 impl<'a> Iterator for RawCertParser<'a>
 {
     type Item = Result<RawCert<'a>>;
@@ -758,7 +795,7 @@ impl<'a> Iterator for RawCertParser<'a>
 
         // (Tag, header length, offset from start of the certificate)
         let mut packets: Vec<(Tag, usize, usize)> = Vec::new();
-        let mut fingerprint = None;
+        let mut primary_key = None;
 
         let mut pending_error = None;
         'packet_parser: loop {
@@ -830,7 +867,7 @@ impl<'a> Iterator for RawCertParser<'a>
                 // Fabricate a header.
                 t!("Recovered after {} bytes of junk", skip);
 
-                pending_error = Some(Error::MalformedPacket(
+                pending_error = Some(crate::Error::MalformedPacket(
                     format!("Encountered {} bytes of junk at offset {}",
                             skip, self.bytes_read)).into());
 
@@ -886,12 +923,11 @@ impl<'a> Iterator for RawCertParser<'a>
                                     Err(err) => {
                                         t!("Stopping: parsing public key: {}",
                                            err);
-                                        pending_error = Some(err);
-                                        break;
-                                    }
-                                    Ok(key) => {
-                                        fingerprint = Some(key.fingerprint());
+                                        primary_key = Some(Err(err));
                                     }
+                                    Ok(key) => primary_key = Some(
+                                        Ok(key.parts_into_public()
+                                           .role_into_primary())),
                                 }
                             }
                         }
@@ -902,7 +938,7 @@ impl<'a> Iterator for RawCertParser<'a>
                         for {} packets",
                        tag);
                     pending_error = Some(
-                        Error::MalformedPacket(
+                        crate::Error::MalformedPacket(
                             format!("Packet {} uses partial body length \
                                      encoding, which is not allowed in \
                                      certificates",
@@ -916,7 +952,7 @@ impl<'a> Iterator for RawCertParser<'a>
                         for {} packets",
                        tag);
                     pending_error = Some(
-                        Error::MalformedPacket(
+                        crate::Error::MalformedPacket(
                             format!("Packet {} uses intedeterminite length \
                                      encoding, which is not allowed in \
                                      certificates",
@@ -995,21 +1031,38 @@ impl<'a> Iterator for RawCertParser<'a>
             return None;
         }
 
-        Some(Ok(RawCert {
-            data: if let Some(slice) = self.slice.as_ref() {
-                let data = &slice[cert_start_absolute + cert_start
-                                  ..cert_start_absolute + cert_end];
-                assert_eq!(data, cert_data);
-                Cow::Borrowed(data)
-            } else {
-                Cow::Owned(cert_data.to_vec())
-            },
-            fingerprint: fingerprint.expect("set"),
-            packets,
-        }))
+        match primary_key.expect("set") {
+            Ok(primary_key) => Some(Ok(RawCert {
+                data: if let Some(slice) = self.slice.as_ref() {
+                    let data = &slice[cert_start_absolute + cert_start
+                                      ..cert_start_absolute + cert_end];
+                    assert_eq!(data, cert_data);
+                    Cow::Borrowed(data)
+                } else {
+                    Cow::Owned(cert_data.to_vec())
+                },
+                primary_key,
+                packets,
+            })),
+            Err(err) =>
+                Some(Err(Error::UnsupportedCert(err, cert_data.into()).into())),
+        }
     }
 }
 
+/// Errors used in this module.
+#[non_exhaustive]
+#[derive(thiserror::Error, Debug)]
+pub enum Error {
+    /// Unsupported Cert.
+    ///
+    /// This usually occurs, because the primary key is in an
+    /// unsupported format.  In particular, Sequoia does not support
+    /// version 3 keys.
+    #[error("Unsupported Cert: {0}")]
+    UnsupportedCert(anyhow::Error, Vec<u8>),
+}
+
 #[cfg(test)]
 mod test {
     use super::*;
@@ -1030,8 +1083,8 @@ mod test {
             return;
         }
 
-        let a: Vec<Packet> = a.into();
-        let b: Vec<Packet> = b.into();
+        let a = a.into_tsk().into_packets().collect::<Vec<_>>();
+        let b = b.into_tsk().into_packets().collect::<Vec<_>>();
 
         for (i, (a, b)) in a.iter().zip(b.iter()).enumerate() {
             if a != b {
@@ -1114,13 +1167,13 @@ mod test {
                 assert_eq!(c.fingerprint(), r.fingerprint());
 
                 eprintln!("CertParser says:");
-                for (i, p) in c.clone().into_iter().enumerate() {
+                for (i, p) in c.clone().into_tsk().into_packets().enumerate() {
                     eprintln!("  - {}. {}", i, p.tag());
                 }
 
                 let rp = Cert::from_bytes(r.as_bytes()).unwrap();
                 eprintln!("RawCertParser says:");
-                for (i, p) in rp.clone().into_iter().enumerate() {
+                for (i, p) in rp.clone().into_tsk().into_packets().enumerate() {
                     eprintln!("  - {}. {}", i, p.tag());
                 }
 
@@ -1225,9 +1278,9 @@ mod test {
         tracer!(TRACE, "invalid_packets", 0);
 
         let (cert, _) =
-            CertBuilder::general_purpose(None, Some("alice@example.org"))
+            CertBuilder::general_purpose(Some("alice@example.org"))
             .generate()?;
-        let cert : Vec<Packet> = cert.into();
+        let cert = cert.into_packets().collect::<Vec<_>>();
 
         // A userid packet.
         let userid : Packet = cert.clone()
@@ -1239,11 +1292,11 @@ mod test {
         // An unknown packet.
         let tag = Tag::Private(61);
         let unknown : Packet
-            = Unknown::new(tag, Error::UnsupportedPacketType(tag).into())
+            = Unknown::new(tag, crate::Error::UnsupportedPacketType(tag).into())
             .into();
 
         // A literal packet.  (This is a valid OpenPGP Message.)
-        let mut lit = Literal::new(DataFormat::Text);
+        let mut lit = Literal::new(DataFormat::Unicode);
         lit.set_body(b"test".to_vec());
         let lit = Packet::from(lit);
 
@@ -1405,14 +1458,14 @@ mod test {
         for i in 0..n {
             let (cert, _) =
                 CertBuilder::general_purpose(
-                    None, Some(format!("{}@example.org", i)))
+                    Some(format!("{}@example.org", i)))
                 .generate()?;
 
             cert.as_tsk().serialize(&mut data)?;
             certs_orig.push(cert);
 
             if literal {
-                let mut lit = Literal::new(DataFormat::Text);
+                let mut lit = Literal::new(DataFormat::Unicode);
                 lit.set_body(b"data".to_vec());
 
                 Packet::from(lit).serialize(&mut data)?;
@@ -1513,14 +1566,14 @@ mod test {
 
         let (cert_1, _) =
             CertBuilder::general_purpose(
-                None, Some("a@example.org"))
+                Some("a@example.org"))
             .generate()?;
         let cert_1_packets: Vec<Packet>
-            = cert_1.into_packets2().collect();
+            = cert_1.into_packets().collect();
 
         let (cert_2, _) =
             CertBuilder::general_purpose(
-                None, Some("b@example.org"))
+                Some("b@example.org"))
             .generate()?;
 
         for n in 1..cert_1_packets.len() {
@@ -1588,7 +1641,7 @@ mod test {
 
         assert_eq!(
             raw.keys().map(|k| k.fingerprint()).collect::<Vec<_>>(),
-            cert.keys().map(|k| k.fingerprint()).collect::<Vec<_>>());
+            cert.keys().map(|k| k.key().fingerprint()).collect::<Vec<_>>());
 
         assert_eq!(
             raw.userids().collect::<Vec<_>>(),
@@ -1608,11 +1661,11 @@ mod test {
         }
 
         assert!(
-            matches!(cert.map_err(|e| e.downcast::<Error>()),
-                     Err(Ok(Error::MalformedCert(_)))));
+            matches!(cert.map_err(|e| e.downcast::<crate::Error>()),
+                     Err(Ok(crate::Error::MalformedCert(_)))));
         assert!(
-            matches!(raw.map_err(|e| e.downcast::<Error>()),
-                     Err(Ok(Error::MalformedCert(_)))));
+            matches!(raw.map_err(|e| e.downcast::<crate::Error>()),
+                     Err(Ok(crate::Error::MalformedCert(_)))));
 
         // Parse two certificates.
         let mut bytes = Vec::new();
@@ -1635,11 +1688,11 @@ mod test {
         }
 
         assert!(
-            matches!(cert.map_err(|e| e.downcast::<Error>()),
-                     Err(Ok(Error::MalformedCert(_)))));
+            matches!(cert.map_err(|e| e.downcast::<crate::Error>()),
+                     Err(Ok(crate::Error::MalformedCert(_)))));
         assert!(
-            matches!(raw.map_err(|e| e.downcast::<Error>()),
-                     Err(Ok(Error::MalformedCert(_)))));
+            matches!(raw.map_err(|e| e.downcast::<crate::Error>()),
+                     Err(Ok(crate::Error::MalformedCert(_)))));
     }
 
     #[test]
