@@ -3,7 +3,12 @@
 use crate::packet::{self, key, Key};
 use crate::crypto::SessionKey;
 use crate::crypto::mpi;
-use crate::types::{Curve, HashAlgorithm, PublicKeyAlgorithm};
+use crate::types::{
+    Curve,
+    HashAlgorithm,
+    PublicKeyAlgorithm,
+    SymmetricAlgorithm,
+};
 
 use crate::{Error, Result};
 
@@ -49,7 +54,7 @@ pub trait Signer {
     /// be used. Provide an explicit implementation only when a smaller subset
     /// of hashing algorithms is valid for this `Signer` implementation.
     fn acceptable_hashes(&self) -> &[HashAlgorithm] {
-        &crate::crypto::hash::DEFAULT_HASHES_SORTED
+        crate::crypto::hash::default_hashes_sorted()
     }
 
     /// Creates a signature over the `digest` produced by `hash_algo`.
@@ -166,7 +171,7 @@ impl Decryptor for Box<dyn Decryptor + Send + Sync> {
 ///
 /// // Or from a query over a certificate:
 /// let (cert, _) =
-///     CertBuilder::general_purpose(None, Some("alice@example.org"))
+///     CertBuilder::general_purpose(Some("alice@example.org"))
 ///         .generate()?;
 /// let keypair =
 ///     cert.keys().unencrypted_secret().nth(0).unwrap().key().clone()
@@ -221,7 +226,24 @@ impl Signer for KeyPair {
         use crate::crypto::backend::{Backend, interface::Asymmetric};
 
         self.secret().map(|secret| {
+            #[allow(deprecated)]
             match (self.public().pk_algo(), self.public().mpis(), secret) {
+                (PublicKeyAlgorithm::Ed25519,
+                 mpi::PublicKey::Ed25519 { a },
+                 mpi::SecretKeyMaterial::Ed25519 { x }) => {
+                    Ok(mpi::Signature::Ed25519 {
+                        s: Box::new(Backend::ed25519_sign(x, a, digest)?),
+                    })
+                },
+
+                (PublicKeyAlgorithm::Ed448,
+                 mpi::PublicKey::Ed448 { a },
+                 mpi::SecretKeyMaterial::Ed448 { x }) => {
+                    Ok(mpi::Signature::Ed448 {
+                        s: Box::new(Backend::ed448_sign(x, a, digest)?),
+                    })
+                },
+
                 (PublicKeyAlgorithm::EdDSA,
                  mpi::PublicKey::EdDSA { curve, q },
                  mpi::SecretKeyMaterial::EdDSA { scalar }) => match curve {
@@ -240,6 +262,13 @@ impl Signer for KeyPair {
                         Error::UnsupportedEllipticCurve(curve.clone()).into()),
                 },
 
+                (PublicKeyAlgorithm::DSA,
+                 mpi::PublicKey::DSA { p, q, g, y },
+                 mpi::SecretKeyMaterial::DSA { x }) => {
+                    let (r, s) = Backend::dsa_sign(x, p, q, g, y, digest)?;
+                    Ok(mpi::Signature::DSA { r, s })
+                },
+
                 (_algo, _public, secret) =>
                     self.sign_backend(secret, hash_algo, digest),
             }
@@ -257,11 +286,60 @@ impl Decryptor for KeyPair {
                plaintext_len: Option<usize>)
                -> Result<SessionKey>
     {
-        use crate::crypto::backend::{Backend, interface::Asymmetric};
+        use crate::crypto::ecdh::aes_key_unwrap;
+        use crate::crypto::backend::{Backend, interface::{Asymmetric, Kdf}};
 
         self.secret().map(|secret| {
             #[allow(non_snake_case)]
             match (self.public().mpis(), secret, ciphertext) {
+                (mpi::PublicKey::X25519 { u: U },
+                 mpi::SecretKeyMaterial::X25519 { x },
+                 mpi::Ciphertext::X25519 { e: E, key }) => {
+                    // Compute the shared point S = xE;
+                    let S = Backend::x25519_shared_point(x, E)?;
+
+                    // Compute the wrap key.
+                    let wrap_algo = SymmetricAlgorithm::AES128;
+                    let mut ikm: SessionKey = vec![0; 32 + 32 + 32].into();
+
+                    // Yes clippy, this operation will always return
+                    // zero.  This is the intended outcome.  Chill.
+                    #[allow(clippy::erasing_op)]
+                    ikm[0 * 32..1 * 32].copy_from_slice(&E[..]);
+                    ikm[1 * 32..2 * 32].copy_from_slice(&U[..]);
+                    ikm[2 * 32..3 * 32].copy_from_slice(&S[..]);
+                    let mut kek = vec![0; wrap_algo.key_size()?].into();
+                    Backend::hkdf_sha256(&ikm, None, b"OpenPGP X25519",
+                                         &mut kek)?;
+
+                    Ok(aes_key_unwrap(wrap_algo, kek.as_protected(),
+                                      key)?.into())
+                },
+
+                (mpi::PublicKey::X448 { u: U },
+                 mpi::SecretKeyMaterial::X448 { x },
+                 mpi::Ciphertext::X448 { e: E, key }) => {
+                    // Compute the shared point S = xE;
+                    let S = Backend::x448_shared_point(x, E)?;
+
+                    // Compute the wrap key.
+                    let wrap_algo = SymmetricAlgorithm::AES256;
+                    let mut ikm: SessionKey = vec![0; 56 + 56 + 56].into();
+
+                    // Yes clippy, this operation will always return
+                    // zero.  This is the intended outcome.  Chill.
+                    #[allow(clippy::erasing_op)]
+                    ikm[0 * 56..1 * 56].copy_from_slice(&E[..]);
+                    ikm[1 * 56..2 * 56].copy_from_slice(&U[..]);
+                    ikm[2 * 56..3 * 56].copy_from_slice(&S[..]);
+                    let mut kek = vec![0; wrap_algo.key_size()?].into();
+                    Backend::hkdf_sha512(&ikm, None, b"OpenPGP X448",
+                                         &mut kek)?;
+
+                    Ok(aes_key_unwrap(wrap_algo, kek.as_protected(),
+                                      key)?.into())
+                },
+
                 (mpi::PublicKey::ECDH { curve: Curve::Cv25519, .. },
                  mpi::SecretKeyMaterial::ECDH { scalar, },
                  mpi::Ciphertext::ECDH { e, .. }) =>
@@ -283,7 +361,7 @@ impl Decryptor for KeyPair {
                     // (r, R) is the recipient's key pair.
                     let S = Backend::x25519_shared_point(&r, &V.try_into()?)?;
 
-                    crate::crypto::ecdh::decrypt_unwrap2(
+                    crate::crypto::ecdh::decrypt_unwrap(
                         self.public(), &S, ciphertext, plaintext_len)
                 },
 
