@@ -28,6 +28,8 @@ impl Asymmetric for super::Backend {
         use PublicKeyAlgorithm::*;
         #[allow(deprecated)]
         match algo {
+            X25519 | Ed25519 |
+            X448 | Ed448 |
             RSAEncryptSign | RSAEncrypt | RSASign => true,
             DSA => true,
             ECDH | ECDSA | EdDSA => true,
@@ -76,6 +78,30 @@ impl Asymmetric for super::Backend {
         Ok(deriver.derive_to_vec()?.into())
     }
 
+    fn x448_generate_key() -> Result<(Protected, [u8; 56])> {
+        let pair = openssl::pkey::PKey::generate_x448()?;
+        Ok((pair.raw_private_key()?.into(),
+            pair.raw_public_key()?.as_slice().try_into()?))
+    }
+
+    fn x448_derive_public(secret: &Protected) -> Result<[u8; 56]> {
+        let key = PKey::private_key_from_raw_bytes(
+            secret, openssl::pkey::Id::X448)?;
+        Ok(key.raw_public_key()?.as_slice().try_into()?)
+    }
+
+    fn x448_shared_point(secret: &Protected, public: &[u8; 56])
+                           -> Result<Protected> {
+        let public = PKey::public_key_from_raw_bytes(
+            public, openssl::pkey::Id::X448)?;
+        let secret = PKey::private_key_from_raw_bytes(
+            secret, openssl::pkey::Id::X448)?;
+
+        let mut deriver = Deriver::new(&secret)?;
+        deriver.set_peer(&public)?;
+        Ok(deriver.derive_to_vec()?.into())
+    }
+
     fn ed25519_generate_key() -> Result<(Protected, [u8; 32])> {
         let pair = openssl::pkey::PKey::generate_ed25519()?;
         Ok((pair.raw_private_key()?.into(),
@@ -105,6 +131,35 @@ impl Asymmetric for super::Backend {
         Ok(verifier.verify_oneshot(signature, digest)?)
     }
 
+    fn ed448_generate_key() -> Result<(Protected, [u8; 57])> {
+        let pair = openssl::pkey::PKey::generate_ed448()?;
+        Ok((pair.raw_private_key()?.into(),
+            pair.raw_public_key()?.as_slice().try_into()?))
+    }
+
+    fn ed448_derive_public(secret: &Protected) -> Result<[u8; 57]> {
+        let key = PKey::private_key_from_raw_bytes(
+            secret, openssl::pkey::Id::ED448)?;
+        Ok(key.raw_public_key()?.as_slice().try_into()?)
+    }
+
+    fn ed448_sign(secret: &Protected, _public: &[u8; 57], digest: &[u8])
+                    -> Result<[u8; 114]> {
+        let key = PKey::private_key_from_raw_bytes(
+            secret, openssl::pkey::Id::ED448)?;
+
+        let mut signer = OpenSslSigner::new_without_digest(&key)?;
+        Ok(signer.sign_oneshot_to_vec(digest)?.as_slice().try_into()?)
+    }
+
+    fn ed448_verify(public: &[u8; 57], digest: &[u8], signature: &[u8; 114])
+                      -> Result<bool> {
+        let key = PKey::public_key_from_raw_bytes(
+            public, openssl::pkey::Id::ED448)?;
+        let mut verifier = Verifier::new_without_digest(&key)?;
+        Ok(verifier.verify_oneshot(signature, digest)?)
+    }
+
     fn dsa_generate_key(p_bits: usize)
                         -> Result<(MPI, MPI, MPI, MPI, ProtectedMPI)>
     {
@@ -113,6 +168,49 @@ impl Asymmetric for super::Backend {
         Ok((key.p().into(), key.q().into(), key.g().into(),
             key.pub_key().into(), key.priv_key().into()))
     }
+
+    fn dsa_sign(x: &ProtectedMPI,
+                p: &MPI, q: &MPI, g: &MPI, y: &MPI,
+                digest: &[u8])
+                -> Result<(MPI, MPI)>
+    {
+        use openssl::dsa::{Dsa, DsaSig};
+        let dsa = Dsa::from_private_components(
+            p.try_into()?,
+            q.try_into()?,
+            g.try_into()?,
+            x.try_into()?,
+            y.try_into()?,
+        )?;
+        let key: PKey<_> = dsa.try_into()?;
+        let mut ctx = PkeyCtx::new(&key)?;
+        ctx.sign_init()?;
+        let mut signature = vec![];
+        ctx.sign_to_vec(&digest, &mut signature)?;
+        let signature = DsaSig::from_der(&signature)?;
+        Ok((signature.r().to_vec().into(), signature.s().to_vec().into()))
+    }
+
+    fn dsa_verify(p: &MPI, q: &MPI, g: &MPI, y: &MPI,
+                  digest: &[u8],
+                  r: &MPI, s: &MPI)
+                  -> Result<bool>
+    {
+        use openssl::dsa::{Dsa, DsaSig};
+        let dsa = Dsa::from_public_components(
+            p.try_into()?,
+            q.try_into()?,
+            g.try_into()?,
+            y.try_into()?,
+        )?;
+        let key: PKey<_> = dsa.try_into()?;
+        let r = r.try_into()?;
+        let s = s.try_into()?;
+        let signature = DsaSig::from_private_components(r, s)?;
+        let mut ctx = PkeyCtx::new(&key)?;
+        ctx.verify_init()?;
+        Ok(ctx.verify(&digest, &signature.to_der()?)?)
+    }
 }
 
 impl TryFrom<&ProtectedMPI> for BigNum {
@@ -163,7 +261,7 @@ impl TryFrom<&Curve> for Nid {
             Curve::NistP384 => Nid::SECP384R1,
             Curve::NistP521 => Nid::SECP521R1,
             Curve::BrainpoolP256 => Nid::BRAINPOOL_P256R1,
-            Curve::Unknown(_) if curve.is_brainpoolp384() =>  Nid::BRAINPOOL_P384R1,
+            Curve::BrainpoolP384 => Nid::BRAINPOOL_P384R1,
             Curve::BrainpoolP512 => Nid::BRAINPOOL_P512R1,
             Curve::Ed25519 | // Handled differently.
             Curve::Cv25519 | // Handled differently.
@@ -215,31 +313,7 @@ impl KeyPair {
                         s: signature.into(),
                     })
                 }
-                (
-                    PublicKeyAlgorithm::DSA,
-                    mpi::PublicKey::DSA { p, q, g, y },
-                    mpi::SecretKeyMaterial::DSA { x },
-                ) => {
-                    use openssl::dsa::{Dsa, DsaSig};
-                    let dsa = Dsa::from_private_components(
-                        p.try_into()?,
-                        q.try_into()?,
-                        g.try_into()?,
-                        x.try_into()?,
-                        y.try_into()?,
-                    )?;
-                    let key: PKey<_> = dsa.try_into()?;
-                    let mut ctx = PkeyCtx::new(&key)?;
-                    ctx.sign_init()?;
-                    let mut signature = vec![];
-                    ctx.sign_to_vec(&digest, &mut signature)?;
-                    let signature = DsaSig::from_der(&signature)?;
 
-                    Ok(mpi::Signature::DSA {
-                        r: signature.r().to_vec().into(),
-                        s: signature.s().to_vec().into(),
-                    })
-                }
                 (
                     PublicKeyAlgorithm::ECDSA,
                     mpi::PublicKey::ECDSA { curve, q },
@@ -358,12 +432,13 @@ impl<P: key::KeyParts, R: key::KeyRole> Key<P, R> {
 
             ECDH => crate::crypto::ecdh::encrypt(self.parts_as_public(), data),
 
-            RSASign | DSA | ECDSA | EdDSA =>
+            RSASign | DSA | ECDSA | EdDSA | Ed25519 | Ed448 =>
                 Err(Error::InvalidOperation(
                     format!("{} is not an encryption algorithm", self.pk_algo())
                 ).into()),
 
             ElGamalEncrypt | ElGamalEncryptSign |
+            X25519 | X448 |
             Private(_) | Unknown(_) =>
                 Err(Error::UnsupportedPublicKeyAlgorithm(self.pk_algo()).into()),
         }
@@ -392,22 +467,7 @@ impl<P: key::KeyParts, R: key::KeyRole> Key<P, R> {
                 ctx.verify_init()?;
                 ctx.verify(&v, signature)?
             }
-            (mpi::PublicKey::DSA { p, q, g, y }, mpi::Signature::DSA { r, s }) => {
-                use openssl::dsa::{Dsa, DsaSig};
-                let dsa = Dsa::from_public_components(
-                    p.try_into()?,
-                    q.try_into()?,
-                    g.try_into()?,
-                    y.try_into()?,
-                )?;
-                let key: PKey<_> = dsa.try_into()?;
-                let r = r.try_into()?;
-                let s = s.try_into()?;
-                let signature = DsaSig::from_private_components(r, s)?;
-                let mut ctx = PkeyCtx::new(&key)?;
-                ctx.verify_init()?;
-                ctx.verify(&digest, &signature.to_der()?)?
-            }
+
             (mpi::PublicKey::ECDSA { curve, q }, mpi::Signature::ECDSA { s, r }) => {
                 let nid = curve.try_into()?;
                 let group = EcGroup::from_curve_name(nid)?;
@@ -442,11 +502,11 @@ impl<R> Key4<SecretParts, R>
 where
     R: key::KeyRole,
 {
-    /// Creates a new OpenPGP public key packet for an existing RSA key.
+    /// Creates a new OpenPGP secret key packet for an existing RSA key.
     ///
-    /// The RSA key will use public exponent `e` and modulo `n`. The key will
-    /// have it's creation date set to `ctime` or the current time if `None`
-    /// is given.
+    /// The RSA key will use the secret exponent `d`, derived from the
+    /// secret primes `p` and `q`.  The key will have its creation
+    /// date set to `ctime` or the current time if `None` is given.
     pub fn import_secret_rsa<T>(d: &[u8], p: &[u8], q: &[u8], ctime: T) -> Result<Self>
     where
         T: Into<Option<SystemTime>>,
