@@ -11,22 +11,24 @@ use crate::Result;
 
 /// A short identifier for certificates and keys.
 ///
-/// A `KeyID` identifies a public key.  It is used, for example, to
-/// reference the issuing key of a signature in its [`Issuer`]
-/// subpacket.
+/// A `KeyID` identifies a public key.  It was used in [RFC 4880]: for
+/// example to reference the issuing key of a signature in its
+/// [`Issuer`] subpacket.  You should prefer [`Fingerprint`] over
+/// [`KeyID`] in data structures, interfaces, and wire formats, unless
+/// space is of the utmost concern.
 ///
-/// Currently, Sequoia supports *version 4* fingerprints and Key IDs
-/// only.  *Version 3* fingerprints and Key IDs were deprecated by
-/// [RFC 4880] in 2007.
+/// Currently, Sequoia supports *version 6* fingerprints and Key IDs,
+/// and *version 4* fingerprints and Key IDs.  *Version 3*
+/// fingerprints and Key IDs were deprecated by [RFC 4880] in 2007.
 ///
-/// A *v4* `KeyID` is defined as a fragment (the lower 8 bytes) of the
-/// key's fingerprint, which in turn is essentially a SHA-1 hash of
-/// the public key packet.  As a general rule of thumb, you should
-/// prefer the fingerprint as it is possible to create keys with a
-/// colliding KeyID using a [birthday attack].
+/// *Version 6* and *version 4* [`KeyID`]s are a truncated version of
+/// the key's fingerprint, which in turn is hash of the public key
+/// packet.  As a general rule of thumb, you should prefer the
+/// fingerprint as it is possible to create keys with a colliding
+/// KeyID using a [birthday attack].
 ///
 /// For more details about how a `KeyID` is generated, see [Section
-/// 12.2 of RFC 4880].
+/// 5.5.4 of RFC 9580].
 ///
 /// In previous versions of OpenPGP, the Key ID used to be called
 /// "long Key ID", as there even was a "short Key ID". At only 4 bytes
@@ -37,15 +39,12 @@ use crate::Result;
 /// See also [`Fingerprint`] and [`KeyHandle`].
 ///
 ///   [RFC 4880]: https://tools.ietf.org/html/rfc4880
-///   [Section 12.2 of RFC 4880]: https://tools.ietf.org/html/rfc4880#section-12.2
+///   [Section 5.5.4 of RFC 9580]: https://www.rfc-editor.org/rfc/rfc9580.html#section-5.5.4
 ///   [birthday attack]: https://nullprogram.com/blog/2019/07/22/
 ///   [`Issuer`]: crate::packet::signature::subpacket::SubpacketValue::Issuer
 ///   [`Fingerprint`]: crate::Fingerprint
 ///   [`KeyHandle`]: crate::KeyHandle
 ///
-/// Note: This enum cannot be exhaustively matched to allow future
-/// extensions.
-///
 /// # Examples
 ///
 /// ```rust
@@ -61,8 +60,12 @@ use crate::Result;
 #[non_exhaustive]
 #[derive(PartialEq, Eq, PartialOrd, Ord, Clone, Hash)]
 pub enum KeyID {
-    /// Lower 8 byte SHA-1 hash.
-    V4([u8;8]),
+    /// A long (8 bytes) key ID.
+    ///
+    /// For v4, this is the right-most 8 bytes of the v4 fingerprint.
+    /// For v6, this is the left-most 8 bytes of the v6 fingerprint.
+    Long([u8; 8]),
+
     /// Used for holding invalid keyids encountered during parsing
     /// e.g. wrong number of bytes.
     Invalid(Box<[u8]>),
@@ -123,7 +126,7 @@ impl From<KeyID> for Vec<u8> {
     fn from(id: KeyID) -> Self {
         let mut r = Vec::with_capacity(8);
         match id {
-            KeyID::V4(ref b) => r.extend_from_slice(b),
+            KeyID::Long(ref b) => r.extend_from_slice(b),
             KeyID::Invalid(ref b) => r.extend_from_slice(b),
         }
         r
@@ -147,10 +150,10 @@ impl From<&Fingerprint> for KeyID {
         match fp {
             Fingerprint::V4(fp) =>
                 KeyID::from_bytes(&fp[fp.len() - 8..]),
-            Fingerprint::V5(fp) =>
-                KeyID::Invalid(fp.iter().cloned().collect()),
-            Fingerprint::Invalid(fp) => {
-                KeyID::Invalid(fp.clone())
+            Fingerprint::V6(fp) =>
+                KeyID::from_bytes(&fp[..8]),
+            Fingerprint::Unknown { bytes, .. } => {
+                KeyID::Invalid(bytes.clone())
             }
         }
     }
@@ -161,10 +164,10 @@ impl From<Fingerprint> for KeyID {
         match fp {
             Fingerprint::V4(fp) =>
                 KeyID::from_bytes(&fp[fp.len() - 8..]),
-            Fingerprint::V5(fp) =>
-                KeyID::Invalid(fp.into()),
-            Fingerprint::Invalid(fp) => {
-                KeyID::Invalid(fp)
+            Fingerprint::V6(fp) =>
+                KeyID::from_bytes(&fp[..8]),
+            Fingerprint::Unknown { bytes, .. } => {
+                KeyID::Invalid(bytes)
             }
         }
     }
@@ -202,7 +205,7 @@ impl KeyID {
     /// ```
     pub fn as_u64(&self) -> Result<u64> {
         match &self {
-            KeyID::V4(ref b) =>
+            KeyID::Long(ref b) =>
                 Ok(u64::from_be_bytes(*b)),
             KeyID::Invalid(_) =>
                 Err(Error::InvalidArgument("Invalid KeyID".into()).into()),
@@ -228,7 +231,7 @@ impl KeyID {
         if raw.len() == 8 {
             let mut keyid : [u8; 8] = Default::default();
             keyid.copy_from_slice(raw);
-            KeyID::V4(keyid)
+            KeyID::Long(keyid)
         } else {
             KeyID::Invalid(raw.to_vec().into_boxed_slice())
         }
@@ -251,16 +254,16 @@ impl KeyID {
     /// ```
     pub fn as_bytes(&self) -> &[u8] {
         match self {
-            KeyID::V4(ref id) => id,
+            KeyID::Long(ref id) => id,
             KeyID::Invalid(ref id) => id,
         }
     }
 
     /// Creates a wildcard `KeyID`.
     ///
-    /// Refer to [Section 5.1 of RFC 4880] for details.
+    /// Refer to [Section 5.1 of RFC 9580] for details.
     ///
-    ///   [Section 5.1 of RFC 4880]: https://tools.ietf.org/html/rfc4880#section-5.1
+    ///   [Section 5.1 of RFC 9580]: https://www.rfc-editor.org/rfc/rfc9580.html#section-5.1
     ///
     /// # Examples
     ///
@@ -276,9 +279,9 @@ impl KeyID {
 
     /// Returns `true` if this is the wildcard `KeyID`.
     ///
-    /// Refer to [Section 5.1 of RFC 4880] for details.
+    /// Refer to [Section 5.1 of RFC 9580] for details.
     ///
-    ///   [Section 5.1 of RFC 4880]: https://tools.ietf.org/html/rfc4880#section-5.1
+    ///   [Section 5.1 of RFC 9580]: https://www.rfc-editor.org/rfc/rfc9580.html#section-5.1
     ///
     /// # Examples
     ///
@@ -394,11 +397,11 @@ impl KeyID {
         let pretty = f.alternate();
 
         let raw = match self {
-            KeyID::V4(ref fp) => &fp[..],
+            KeyID::Long(ref fp) => &fp[..],
             KeyID::Invalid(ref fp) => &fp[..],
         };
 
-        // We currently only handle V4 Key IDs, which look like:
+        // We currently only handle long Key IDs, which look like:
         //
         //   AACB 3243 6300 52D9
         //
@@ -488,7 +491,7 @@ impl KeyID {
             (k, KeyHandle::KeyID(o)) => {
                 k == o
             },
-            (KeyID::V4(k), KeyHandle::Fingerprint(Fingerprint::V4(o))) => {
+            (KeyID::Long(k), KeyHandle::Fingerprint(Fingerprint::V4(o))) => {
                 // Avoid a heap allocation by embedding our
                 // knowledge of how a v4 key ID is derived from a
                 // v4 fingerprint:
@@ -497,6 +500,13 @@ impl KeyID {
                 // fingerprint.
                 &o[12..] == k
             },
+
+            (KeyID::Long(k), KeyHandle::Fingerprint(Fingerprint::V6(f))) => {
+                // A v6 key ID are the 8 left-most octets of a v6
+                // fingerprint.
+                k == &f[..8]
+            },
+
             (k, o) => {
                 k == &KeyID::from(o)
             },
